/**************************************************************************
 *         Machine Intelligence Coding / Cracking, 1997-98                *
 *                      Breaking it open wide.                            *
 **************************************************************************/

#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

//
//  The basic use of this program:
//
//       Crack a program as you normally would, finding the bytes to change
//       and changing them, but KEEP A BACKUP OF THE ORIGINAL.  (With Hex
//       Workshop this is easy, simply say 'YES' to make a backup of original
//       when saving.)  I'll assume you're saving them as .BAK.
//
//       After you've patched the necessary files, say:
//                  fc /b *.bak *.exe > MICPatch.pat
//
//       (The FC program comes with DOS, Windows 95, 98, and Windows NT.
//       The order of the parameters (backup first, executable second)
//       is IMPORTANT!)
//
//       Then when you want to apply the patch simply run this program in
//       the same directory as the executable files and MICPatch.pat.  If
//       you want to, you can name the patch file anything you'd like, and
//       say:
//
//                  MICPatch <patchfilename>
//
//       The source is being released, 'cause it's kinda cool, useful, and
//       maybe it'll help someone learn, and then maybe someday THEY'LL
//       crack a program we want.  Invest in the future!
//
//       If you use this in anything, PLEASE GIVE CREDIT TO US.
//
//                       --    Machine Intelligence Coding / Cracking
//

#define DEFAULT_PATCHFILE "MICPatch.pat"
#define COMPARING_STRING "Comparing "
#define FC_MESSAGE "FC: "

#define TRUE 1
#define FALSE 0

char *
GetFileToHack(char *comparing)
{
  char *szAnd;

  szAnd = strstr(comparing, " and ");
  if(szAnd != NULL) {
    szAnd += strlen(" and ");
  }

  return(szAnd);
}

//  Remove terminating CRLF's, spaces, etc.
void
StripEnd(char *szLine)
{
  int i = strlen(szLine) - 1;

  while(i && isspace(szLine[i])) szLine[i--] = 0;
}

//  Seek to the location, check the byte against the
//  original, if it's equal, write the new.
int
ApplyPatch(int iWriteHandle, unsigned char *szPatchInfo, int iCurLine)
{
  char *HexTable = "0123456789ABCDEF";
  unsigned char *szHexBytes;
  unsigned char hexOrig;
  unsigned char hexSrc;
  unsigned char hexDest;
  unsigned long ulSeekLoc;

  szHexBytes = strchr(szPatchInfo, ':');
  if(szHexBytes == NULL) {
    fprintf(stderr, "Patch at line %d in unknown format!\n", iCurLine);
    return(0);
  }

  szHexBytes[0] = 0;  //  Terminate the basic string at the ':'.
  szHexBytes += 2;    //  Increment past the ':' and the initial space.

  ulSeekLoc = strtoul(szPatchInfo, NULL, 16);

  hexSrc = strtoul(szHexBytes, &szHexBytes, 16);

  if(szHexBytes[0] == 0) {
    fprintf(stderr, "Incorrect format of patch on line %d!\n", iCurLine);
    return(0);
  }

  szHexBytes++;

  hexDest= strtoul(szHexBytes, NULL, 16);

  lseek(iWriteHandle, ulSeekLoc, SEEK_SET);

  read(iWriteHandle, &hexOrig, 1);

  if(hexOrig != hexSrc) {
    fprintf(stderr, "Patch at line %d failed to match! (%2.2x/%2.2x/%2.2x)\n", iCurLine, hexOrig, hexSrc, hexDest);
    return(0);
  }

  lseek(iWriteHandle, ulSeekLoc, SEEK_SET);

  write(iWriteHandle, &hexDest, 1);
}

//  Step through the patch file line by line until the end,
//  changing files when it says to, and applying patch values
//  to the current file.
void
WorkOverFile(FILE *fp, char *szFName)
{
  int skipLine;
  int iCurLine = 0;
  char szInLine[256], *szTempFName;
  int iCurPatchHandle;

  while(!feof(fp)) {
    skipLine = FALSE;
    fgets(szInLine, 256, fp);

    if(strlen(szInLine) == 255) { //  Bad line!
      fprintf(stderr, "Line %d in %s too long.\n", iCurLine, szFName);
      return;
    }

    //  Clean off the end of the string.
    StripEnd(szInLine);

    //  If it's an empty line, or an FC message (other than compare)
    //  we want to skip it.
    if(strlen(szInLine) == 0 ||
       strncmp(szInLine, FC_MESSAGE, strlen(FC_MESSAGE)) == 0) {
      skipLine = TRUE;
    }

    if(strncmp(szInLine, COMPARING_STRING, strlen(COMPARING_STRING))==0) {
      if(iCurPatchHandle != -1) {
	close(iCurPatchHandle);
	iCurPatchHandle = -1;
      }

      szTempFName = GetFileToHack(szInLine);

      iCurPatchHandle = open(szTempFName, O_RDWR | O_BINARY);
      if(iCurPatchHandle == -1) {
	fprintf(stderr, "Failed to open %s for patching!\nError: %s\n",
		szTempFName, strerror(errno));
	return;
      } else {
	printf("Patching file %s.\n", szTempFName);
      }
      skipLine = TRUE;
    }

    //  We're PRETTY damn sure it's a patch now...
    if(!skipLine) {
      if(!ApplyPatch(iCurPatchHandle, szInLine, iCurLine)) {
	close(iCurPatchHandle);
	return;
      }
    }

    //  This works around a bug (imnsho) in the libraries that if
    //  we're at the EOF, but we haven't tried to *READ* the EOF
    //  yet, feof(fp) can return false, and we'll try to do a read,
    //  no data will get put into the buffer, and the last line
    //  previously read will be used.  In our case, since we
    //  remove the ':' as the first thing in the ApplyPatch
    //  function, it means we'd get a spurious 'unknown format'
    //  error.
    szInLine[0] = 0;

    iCurLine++;
  }

  if(iCurPatchHandle != -1) {
    close(iCurPatchHandle);
  }

  printf("Patching complete.\n");
}

void
main(int argc, char **argv)
{
  FILE *fp;
  char *szFName;

  printf("M.I.Cracking Patch Tool 1.0, 1997-98 Machine Intelligence Coding / Cracking\n");

  if(argc != 2) {
    szFName = DEFAULT_PATCHFILE;
  } else {
    szFName = argv[1];
  }

  fp = fopen(szFName, "r");

  if(fp == NULL) {
    if(argc != 2) {
      fprintf(stderr, "Usage:\n\tMICPatch <patchfile>\n");
    } else {
      fprintf(stderr, "Cannot open patchfile %s for reading.\n", szFName);
    }
    return;
  }

  WorkOverFile(fp, szFName);

  fclose(fp);
}
